/**
    Copyright (C) 2003  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#ifndef _TC2_AUTOLOAD_H
#define _TC2_AUTOLOAD_H

#if defined __alpha__
#define TC2_AUTOLOAD(fp, ifc, sym)					      \
static void								      \
TC2_AUTOLOAD_FN(fp)(void)						      \
{									      \
    asm volatile ("lda $26, -4($26)\n\t"  /* adjust return address */	      \
		  "stq $26, 0($30)\n\t"					      \
		  "lda $30, -48($30)\n\t" /* save arguments */		      \
		  "stq $16, 0($30)\n\t"					      \
		  "stq $17, 8($30)\n\t"					      \
		  "stq $18, 16($30)\n\t"				      \
		  "stq $19, 24($30)\n\t"				      \
		  "stq $20, 32($30)\n\t"				      \
		  "stq $21, 40($30)"					      \
		  ::);							      \
									      \
    fp = tc2_get_symbol(ifc, sym);					      \
									      \
    asm volatile ("ldq $26, 48($30)\n\t" /* get the jsr insn that called */   \
		  "ldl $2, 0($26)\n\t"   /* us and extract its Rb register */ \
		  "srl $2, 16, $2\n\t"					      \
		  "and $2, 0x1f, $2\n\t"				      \
		  "br $3, 8\n\t"         /* create an insn on the stack to */ \
		  "mov %0, $0\n\t"       /* write the new address to Rb */    \
		  "ret $31, ($26)\n\t"   /* and execute it */		      \
		  "lda $30, -8($30)\n\t"				      \
		  "ldl $4, 0($3)\n\t"					      \
		  "or $4, $2, $4\n\t"					      \
		  "ldl $5, 4($3)\n\t"					      \
		  "stl $4, 0($30)\n\t"					      \
		  "stl $5, 4($30)\n\t"					      \
		  "mb\n\t"						      \
		  "jsr $26, ($30)\n\t"					      \
		  "lda $30, 8($30)\n\t"					      \
		  "ldq $16, 0($30)\n\t"	 /* restore arguments */	      \
		  "ldq $17, 8($30)\n\t"					      \
		  "ldq $18, 16($30)\n\t"				      \
		  "ldq $19, 24($30)\n\t"				      \
		  "ldq $20, 32($30)\n\t"				      \
		  "ldq $21, 40($30)\n\t"				      \
		  "lda $30, 48($30)"					      \
		  :: "r" (fp));						      \
    /* return and reexecute the jsr with the new target */		      \
}
#elif defined __i386__
#define TC2_AUTOLOAD(fp, ifc, sym)					     \
static void								     \
TC2_AUTOLOAD_FN(fp)(void)						     \
{									     \
    fp = tc2_get_symbol(ifc, sym);					     \
    asm volatile ("movl -4(%%ebp), %%ebx\n\t" /* restore ebx */		     \
		  "movl %%ebp, %%esp\n\t"     /* restore stack pointer */    \
		  "popl %%ebp\n\t"            /* and do the same with ebp */ \
		  "jmp *%0"		      /* jump to function */	     \
		  :: "r" (fp));						     \
}
#endif

#ifdef TC2_AUTOLOAD
#define TC2_AUTOLOAD_FN(f) _tc2_autoload_##f
#define TC2_GETSYM(f, ifc, sym) (f##_t) TC2_AUTOLOAD_FN(f)
#else
#warning TC2 symbol autoloading unavailable on your architecture.
#define TC2_AUTOLOAD(fp, ifc, sym)
#define TC2_GETSYM(f, ifc, sym) tc2_get_symbol(ifc, sym)
#endif

#endif
