/**
    Copyright (C) 2001, 2002, 2003, 2004  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#ifndef _KERNEL_H
#define _KERNEL_H

#define TCHASH_NEWAPI

#include <tc2.h>
#include <tclist.h>
#include <tchash.h>

extern tclist_t *module_dirs;
extern int verbose, debug;
extern int module_count;
extern tc2_interface **interfaces;
extern tchash_table_t *avail_syms, *loaded_syms, *modules, *module_names;
extern int unload_modules;

extern int tc2_add_module(char *file, tc2_module_info *md);

/* Loads a module */
extern int tc2_load_module(tc2_module *module, char *opt);

/* Unloads a module */
extern int tc2_unload_module(tc2_module *module);

/* Get module pointer */
extern tc2_module *tc2_get_interface(char *interface);
extern tc2_module *tc2_get_module(char *);

extern int tc2_add_dep(tc2_module *, tc2_module *);
extern int tc2_del_dep(tc2_module *, tc2_module *);
extern int tc2_autoclean(void);
extern int tc2_del_all_deps(tc2_module *);
extern int kreq_init(void);
extern void kreq_free(void);
extern int tc2_init_hook(void);
extern int tc2_init_print(void);
extern void tc2_free_hooks(void);
extern void tc2_free_conf(void);
extern void tc2_free_print(void);
extern void tc2_free_module(void *p);

#define MKSYM(class,sym)					\
    ({ char *buf = malloc(strlen(class) + strlen(sym) + 2);	\
       sprintf(buf,"%s:%s", class, sym); buf; })

#define kprint(tag, level, format, ...)				\
    tc2_print("TC2:" tag, level, format, ## __VA_ARGS__);

#endif
