/**
    Copyright (C) 2001, 2002, 2003, 2004  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <tc2.h>
#include <signal.h>
#include "kernel.h"

#ifdef HAVE_EXECINFO_H
#include <execinfo.h>
#endif

#define OPTSTR "x:c:m:v:s"

int sigtrap = 0;

#ifdef DEBUG
void stacktrace(int sig)
{
    void *stack[32*sizeof(void *)];
    int s;

    psignal(sig, NULL);
#ifdef HAVE_BACKTRACE
    fprintf(stderr, "Stack trace:\n");
    s = backtrace(stack, 32);
    backtrace_symbols_fd(stack, s, 2);
#endif
    while(sigtrap > 1)
	sleep(5);
    abort();
}
#endif

int main(int argc, char **argv)
{
    int i;
    int opt;
    int mcount;
    char **mods;
    char **opts;
    int verbose = 0;
    
    mods = calloc(argc, sizeof(char *));
    opts = calloc(argc, sizeof(char *));

    mcount = 0;
    while((opt = getopt(argc, argv, OPTSTR)) != -1){
	char *mopt;
	switch(opt){
	case 'x':
	    mopt = strchr(optarg, ':');
	    if(mopt != NULL){
		int ml;
		ml = mopt - optarg;
		mods[mcount] = malloc(ml + 1);
		strncpy(mods[mcount], optarg, ml);
		opts[mcount] = strdup(++mopt);
	    } else {
		mods[mcount] = strdup(optarg);
		opts[mcount] = NULL;
	    }
	    mcount++;
	    break;

	case 'c':
	    tc2_add_config(optarg);
	    break;

	case 'm':
	    tc2_add_modpath(optarg);
	    break;

	case 'v':
	    verbose = strtol(optarg, NULL, 0);
	    break;

	case 's':
	    sigtrap++;
#if defined DEBUG && defined HAVE_BACKTRACE
	    if(sigtrap == 1){
		signal(SIGSEGV, stacktrace);
		signal(SIGFPE, stacktrace);
	    }
#endif
	    break;
	}
    }

    tc2_verbose(verbose);

    /* Initialize things. */
    if (tc2_init()) {
	kprint("init", TC2_PRINT_ERROR, "Fatal error initializing.\n");
	exit(1);	
    }

    /* Add the requested modules to queue. */
    for(i = 0; i < mcount; i++){
	kprint("main", TC2_PRINT_DEBUG + 1,
	       "Loading module: %s with option: %s\n", mods[i], opts[i]);
	tc2_request(TC2_LOAD_MODULE, 0, mods[i], opts[i]);
    }

    tc2_run();
    tc2_free();

    for(i = 0; i < mcount; i++){
	free(mods[i]);
	if(opts[i])
	    free(opts[i]);
    }
    free(mods);
    free(opts);

    return 0;
}
